# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import YW_matchups






sat_folder = '/Users/yw/Downloads/test_polymer'
is_file = '/Users/yw/Local_storage/GLORIA/6_output_L8_noRrs.csv'
out_file = '/Users/yw/Downloads/test_polymer/text.csv'






test = YW_matchups.run(sat_folder, is_file, AC='POLYMER', out_file=out_file, to_log = True)






'''
# Plot 

from matplotlib import pyplot as plt
plt.imshow(test3, interpolation='nearest')
plt.show()

'''




# find satellite matchups 

'''
from .Surface import Surface, SpectralSurface
from .Atmosphere import Atmosphere

from .Tmart import Tmart

from .tm_calcref import calc_ref 
from .tm_geometry import dirP_to_coord

from tmart import AEC
from tmart import surface_rho

'''



from .run import *










# Tool package for POLYMER


import netCDF4 as nc4
import ast
import numpy as np
import datetime as dt
import math
import sys



# Get datetime object 
def get_datetime(sat_file):
    with nc4.Dataset(sat_file,"r") as nc:
        
        print('test')
        
        try: 
            datetime_sat_string = nc.getncattr('time_coverage_start')
            datetime_sat_string = datetime_sat_string[:-5]
            format_string = '%Y-%m-%dT%H:%M:%S'
            datetime_sat_obj = dt.datetime.strptime(datetime_sat_string, format_string)
            add_a_day = False
            
        # in case 'time_coverage_start' is missing in the nc file
        except: 
            # product_name = nc.getncattr('product_name')[17:25]
            # format_string = '%Y%m%d'
            # datetime_sat_obj = dt.datetime.strptime(product_name, format_string)   
            # add_a_day = True
            sys.exit('L2gen processing problem')
        
    return [datetime_sat_obj, add_a_day]
    
    

# Read NC data 

def read_file(sat_file):
    
    with nc4.Dataset(sat_file,"r") as nc:
        
        # L2gen outputs use valid_min and valid_max which is problematic sometimes...
        # Disable masks to deal with this 
        nc.set_auto_mask(False)
        
        
        ### Read data 
        
        instrument = nc.getncattr('instrument')
        platform = nc.getncattr('platform')
        
        # wavelengths 
        if instrument == 'MSI':
            column_names = ['B1','B2','B3','B4','B5','B6','B7','B8']
            if platform == 'Sentinel-2A': 
                sensor = 'S2A_MSI'
                bands = [443,492,560,665,704,740,783,835]
                
            elif platform == 'Sentinel-2B': 
                sensor = 'S2B_MSI'
                bands = [442,492,559,665,704,739,780,835]
        
        elif instrument == 'OLI':
            column_names = ['B1','B2','B3','B4']
            bands = [443,482,561,655]
            sensor = 'L8_OLI'
        
        
        # band names for extracting values
        band_names = ['Rrs_' + str(item) for item in bands]

        # dimension
        width = len(nc.dimensions['pixels_per_line']) # equivalent to 'pixels' before 
        height = len(nc.dimensions['number_of_lines']) # equivalent to 'lines' before 

        # latlon
        lat = nc.groups['navigation_data'].variables['latitude'][:,:]
        lon = nc.groups['navigation_data'].variables['longitude'][:,:]

        # to store data
        rhow_data = np.ma.empty([len(bands),int(height),int(width)])

        for k in range(len(bands)):
            varname = band_names[k]
            
            # extract attributes 
            v_attributes = nc.groups['geophysical_data'].variables[varname].__dict__
            v_fill = float(v_attributes['_FillValue'])
            v_scale = float(v_attributes['scale_factor'])
            v_add = float(v_attributes['add_offset'])
            
            # apply mask manually 
            mask_value = (v_fill * v_scale + v_add) * 0.99
            
            arr_temp = nc.groups['geophysical_data'].variables[varname][:,:]
            arr_temp[arr_temp < mask_value] = np.nan
            
            ### Check L8 processing, if the same !!!
            
            rhow_data[k] = arr_temp * math.pi
            
    return {'bands': band_names,            # variable names of the bands 
            'sensor': sensor,
            'column_names': column_names,   # column names for final output
            'width': width,                 # width of the data array
            'height': height,               # height of the data array
            'lat': lat,                     # 2d data array: latitude at [y,x]
            'lon': lon,                     # 2d data array: longitude at [y,x]
            'rhow_data': rhow_data}         # 3d data array: reflectance at [band, y, x]
























# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import YW_matchups




# test Pleiades


sat_folder = '/Volumes/San2T/241027_Pleiades/ACed'



is_file = '/Volumes/San2T/241027_Pleiades/Matchups/lab_for_matchups.csv'
out_file = '/Volumes/San2T/241027_Pleiades/Matchups/Pleiades_lab.csv'


test = YW_matchups.run(sat_folder, is_file, AC='ACOLITE', out_file=out_file, dt_column='datetime',dt_string='%m/%d/%Y %H:%M')











'''
# Plot 

test = test[0]

from matplotlib import pyplot as plt
plt.imshow(test, interpolation='nearest')
plt.show()




plt.imshow(rhow_data[0], interpolation='nearest')
plt.show()


'''




# Go up by 2 directory and import 

import sys

sys.path.append('/Users/yw/Desktop/YW_private_code/YW_matchups')

import YW_matchups


### in situ 
is_file = '/Volumes/San2T/240727_Chapter3/Matchups/spectral.csv'


dt_format = '%Y-%m-%d %H:%M'


# L89
sat_folder = '/Volumes/San2T/240727_Chapter3/ACed/L89'
out_file = '/Users/yw/Downloads/test/spectral_L89.csv'
test = YW_matchups.run(sat_folder, is_file, AC='ACOLITE', out_file=out_file, dt_column='datetime',
                       dt_string=dt_format, ACOLITE_L2R=True)





# Strategy: throw everything here, put them into modules if too long or when necessary 







### Introduction 

# find matchups, mean and median for each of the window sizes 
# sensor in the last column 


### Problem

# POLYMER no sensing time for L8, therefore add a day to the time window 
# should be okay because L8 rarely visits the same place two days in a row





import numpy as np
import sys
import glob, os
import pyproj
import datetime as dt
import pandas as pd
import time
import warnings
from importlib.metadata import version

from . import AC_polymer
from . import AC_acolite
from . import AC_l2gen


# make time column and string an option 



def run(sat_folder=None, is_file=None, AC=None, out_file=None, to_log = True, 
        dt_column='GLORIA_time', dt_string = "%Y-%m-%dT%H:%M", output_sd = False,
        ACOLITE_L2R=False):
    
    
    ### Settings to be included later 
    time_window_minutes = 720 
    max_dist = 100
    
    
    # make it adaptive to sensor resolution ??
    # window_size = 3
    window_size = [1, 3, 5, 7]
    
    
    ### Print all metadata/settings and save them in a txt file
    
    if to_log: 
        # Start logging in txt file
        orig_stdout = sys.stdout
        
        log_file = out_file.replace(".csv", ".txt")
        # log_file = 'tmart_log_' + time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime(time.time())) + '.txt'
    
        class Logger:
            def __init__(self, filename):
                self.console = sys.stdout
                self.file = open(filename, 'w')
                self.file.flush()
            def write(self, message):
                self.console.write(message)
                self.file.write(message)
            def flush(self):
                self.console.flush()
                self.file.flush()
    
        sys.stdout = Logger(log_file)
        
    
    # Metadata
    print('YW_matchups version: ' + str(version('YW_matchups')))
    print('System time: ' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time())))
    print('sat_folder: ' + str(sat_folder))
    print('In-situ file: ' + str(is_file))
    print('AC: ' + str(AC))
    print('out_file: ' + str(out_file))
    print('dt_column: ' + str(dt_column))
    print('dt_string: ' + str(dt_string))
    print('output_sd: ' + str(output_sd))
    print('ACOLITE_L2R: ' + str(ACOLITE_L2R))
    if to_log: print('log_file: ' + str(log_file))
    
    # Settings 
    print('window_size: ' + str(window_size))
    print('time_window_minutes: ' + str(time_window_minutes))
    print('max_dist: ' + str(max_dist))
    
    
    # Read in-situ data 
    df_is = pd.read_csv(is_file)
    dt_is_string = list(df_is.loc[:,dt_column])
    dt_is_obj = [dt.datetime.strptime(dt_is_string[d],dt_string)for d in list(range(len(dt_is_string)))]
    
    
    
    
    ### find all NC files 
    
    # Keywords for .nc files 
    if AC=='POLYMER' or AC=='L2gen': 
        AC_file = '*.nc'
    elif AC=='ACOLITE':
        AC_file = '*L2W.nc'
        if ACOLITE_L2R: AC_file = '*L2R.nc'
    else: 
        sys.exit('Unknown AC algorithm')
    
    # Find them in the directory and subdirectory 
    # https://stackoverflow.com/questions/18394147/how-to-do-a-recursive-sub-folder-search-and-return-files-in-a-list
    sat_files = [y for x in os.walk(sat_folder) for y in glob.glob(os.path.join(x[0], AC_file))]
    
    # Output 
    df_output = pd.DataFrame()
    
    
    ### Loop through each of the sat_files
    
    # sat_i = 2
    # sat_file = sat_files[sat_i]
    
    for sat_i, sat_file in enumerate(sat_files): 
        
        print('\n{}/{}'.format(sat_i, len(sat_files)))
        print('Processing: {}'.format(sat_file))
        
        tw_minutes = time_window_minutes
        
        
        
        # Read datetime, for different AC
        if AC=='POLYMER': 
            dt_sat_obj, add_a_day = AC_polymer.get_datetime(sat_file)
        elif AC=='ACOLITE':
            dt_sat_obj, add_a_day = AC_acolite.get_datetime(sat_file)
        elif AC=='L2gen':
            dt_sat_obj, add_a_day = AC_l2gen.get_datetime(sat_file)

        # add a day for POLYMER L8 because we don't have the sensing time 
        # this is okay because L8 doesn't visit the same place two days in a row
        if add_a_day: tw_minutes = tw_minutes + 1440


        # Indices of in-situ data that match the sat datetime  
        is_matches = [ind for ind, s in enumerate(dt_is_obj) if abs((dt_sat_obj - s).total_seconds()/60) <= tw_minutes]
        

        # If there are no matching records, skip to next satellite file.
        if len(is_matches)==0:
            print("No matches for this file within the time limit. Next...\n\n")
            continue
        elif len(is_matches) > 1:
            print('Number of matchups: ' + str(len(is_matches)))

        ### Read NC data, for different AC
        if AC=='POLYMER': 
            sat_data = AC_polymer.read_file(sat_file)
        elif AC=='ACOLITE':
            sat_data = AC_acolite.read_file(sat_file, ACOLITE_L2R)
        elif AC=='L2gen':
            sat_data = AC_l2gen.read_file(sat_file)
        
        
        ### Loop through indices and row numbers of in-situ matchups
        
        for ind, row in enumerate(is_matches):
            print('Processing in-situ match {}, row {} in spreadsheet'.format(ind, row))
            
            ### Identify distances 
            
            in_situ_lat = df_is.loc[row,"lat"]
            in_situ_lon = df_is.loc[row,"lon"]
            
            sat_lon_max = sat_data['lon'].max()
            sat_lon_min = sat_data['lon'].min()
            sat_lat_max = sat_data['lat'].max()
            sat_lat_min = sat_data['lat'].min()
            
            
            # Make sure in situ is in sat range 
            if in_situ_lat > sat_lat_max or in_situ_lat < sat_lat_min:
                print("Not in the range. Next...\n")
                continue
            if in_situ_lon > sat_lon_max or in_situ_lon < sat_lon_min:
                print("Not in the range. Next...\n")
                continue
            
            
            
            # Project X and Y in meters, centred at the in-situ measurement 
            proj = pyproj.Proj(proj="gnom", ellps="WGS84", lat_0=in_situ_lat,
                               lon_0=in_situ_lon, datum="WGS84", units="m")
            x, y = proj(sat_data['lon'], sat_data['lat'])
            in_situ_x, in_situ_y = proj(in_situ_lon, in_situ_lat)
            
            # Calculate distance and indices of the closest 
            dist = np.sqrt((x - in_situ_x)**2 + (y - in_situ_y)**2)
            
            # Pixels within the max distance, True/False
            within_max_dist = dist < max_dist
            
            # Are there any valid pixels left?
            if np.sum(within_max_dist)==0:
                print("No pixels within acceptable distance (",max_dist,"m ). Next...\n")
                continue
            
            ### Find closest and extract values 
            
            indices_closest = np.argwhere(dist == np.min(dist))
            
            line = indices_closest[0][0]
            pixel = indices_closest[0][1]
            
            # extract the row as a series 
            row_temp = df_is.iloc[row].copy()
            row_temp['image'] = sat_file
            
            # Loop through window size 
            for ws in window_size: 
            
                # Get pixel coordinates of window around the matching point.
                # (v = vertical/line, h = horizontal/pixel)
                v1 = np.int32(line - ((ws-1)/2))
                v2 = np.int32(line + ((ws+1)/2))
                h1 = np.int32(pixel - ((ws-1)/2))
                h2 = np.int32(pixel + ((ws+1)/2))
                
                # Trim edges of window if they're on the edge of the grid.
                line_start,line_end = max(0, v1), min(sat_data['height'], v2)
                # Window pixel start, pixel end.
                pixel_start,pixel_end = max(0, h1), min(sat_data['width'], h2)
                
                # Extract all values in this window
                rhow_data_window = sat_data['rhow_data'][:,line_start:line_end,pixel_start:pixel_end]
                
                # Column names 
                columns_mean    = ['mean_'+item+'_ws'+str(ws) for item in sat_data['column_names']]
                columns_median  = ['median_'+item+'_ws'+str(ws) for item in sat_data['column_names']]
                
                # SD output 
                if output_sd and ws>1: columns_sd  = ['sd_'+item+'_ws'+str(ws) for item in sat_data['column_names']]
                
                
                # Suppress mean/median of empty slice
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                
                    # Mean and median values per band 
                    rhow_data_mean      = np.nanmean(rhow_data_window, axis=(1,2))
                    rhow_data_median    = np.nanmedian(rhow_data_window, axis=(1,2))
                    
                    if output_sd and ws>1: rhow_data_sd = np.nanstd(rhow_data_window, axis=(1,2))
                
                # if rhow_data_mean.mask.all(): 
                #     print('All pixels are masked in the window.')
                #     continue
                
                # Add values to data frame
                row_mean    = pd.Series(data=rhow_data_mean,    index=columns_mean)
                row_median  = pd.Series(data=rhow_data_median,  index=columns_median)
                if output_sd and ws>1: row_sd = pd.Series(data=rhow_data_sd,  index=columns_sd)
                
                
                # if output sd 
                if output_sd and ws>1:
                    row_temp = pd.concat([row_temp, row_mean, row_median, row_sd], axis=0)
                else:
                    row_temp = pd.concat([row_temp, row_mean, row_median], axis=0)

            # Add sensor
            row_temp['sensor'] = sat_data['sensor']

            # Add to dataframe             
            df_output = pd.concat([df_output,row_temp.to_frame().T])
            
    # Stop logging 
    if to_log: sys.stdout = orig_stdout
    
    # Output dataframe 
    df_output.to_csv(out_file, index=False)

    return 0
    
    
    
    
    
    
    
    
    
    
    
    
    
















# Tool package for POLYMER


import netCDF4 as nc4
import ast
import numpy as np
import datetime as dt



# Get datetime object 
def get_datetime(sat_file):
    with nc4.Dataset(sat_file,"r") as nc:
        sensor = nc.getncattr('sensor')
        
        if sensor =='MSI': 
            datetime_sat_string = nc.getncattr('sensing_time')
            format_string = '%Y-%m-%d %H:%M:%S'
            datetime_sat_obj = dt.datetime.strptime(datetime_sat_string, format_string)
            add_a_day = False
            
        elif sensor == 'OLI':
            l1_dirname = nc.getncattr('l1_dirname')
            extracted_date = l1_dirname[-23:-15]
            format_string = '%Y%m%d'
            datetime_sat_obj = dt.datetime.strptime(extracted_date, format_string)   
            add_a_day = True
            
    return [datetime_sat_obj, add_a_day]
    
    

# Read NC data 

def read_file(sat_file):
    
    with nc4.Dataset(sat_file,"r") as nc:
        
        sensor_temp = nc.getncattr('sensor')
        
        if sensor_temp == 'MSI':
            tile_id = nc.getncattr('L1_TILE_ID')
            sensor = tile_id[0:3]
            if sensor == 'S2A' or sensor == 'S2B':
                column_names = ['B1','B2','B3','B4','B5','B6','B7','B8','B8A','B11']
                if sensor == 'S2A': sensor = 'S2A_MSI'
                if sensor == 'S2B': sensor = 'S2B_MSI'
                
        elif sensor_temp == 'OLI':
            sensor = 'L8_OLI'
            column_names = ['B1','B2','B3','B4','B5']
                
        bands = nc.getncattr('bands_rw')
        bands = ast.literal_eval(bands)
        
        # band names for extracting values
        band_names = ['Rw' + str(item) for item in bands]

        # dimension
        width = len(nc.dimensions['width']) # equivalent to 'pixels' before 
        height = len(nc.dimensions['height']) # equivalent to 'lines' before 

        # latlon
        lat = nc.variables['latitude'][:,:]
        lon = nc.variables['longitude'][:,:]

        # to store data
        rhow_data = np.ma.empty([len(bands),int(height),int(width)])

        for k in range(len(bands)):
            varname = band_names[k]
            rhow_data[k] = nc.variables[varname][:,:]
            
    return {'bands': band_names,            # variable names of the bands 
            'sensor': sensor,
            'column_names': column_names,   # column names for final output
            'width': width,                 # width of the data array
            'height': height,               # height of the data array
            'lat': lat,                     # 2d data array: latitude at [y,x]
            'lon': lon,                     # 2d data array: longitude at [y,x]
            'rhow_data': rhow_data}         # 3d data array: reflectance at [band, y, x]
























# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import YW_matchups








sat_folder = '/Users/yw/Downloads/L8_out'

is_file = '/Users/yw/Local_storage/GLORIA/6_output_L8_noRrs.csv'


out_file = '/Users/yw/Downloads/L8_out/text.csv'





test = YW_matchups.run(sat_folder, is_file, AC='L2gen', out_file=out_file, to_log = False)




'''
# Plot 



k=0
varname = band_names[k]
test = nc.groups['geophysical_data'].variables[varname][:]


test = sat_data['rhow_data'][0]



from matplotlib import pyplot as plt
plt.imshow(test, interpolation='nearest')
plt.show()





'''




# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import YW_matchups








# sat_file = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER/test_output_S2_60m.nc'


# sat_folder = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER'

sat_folder = '/Volumes/San2T/GLORIA/ACed/S2_Tmart_ACOLITE/S2A_MSIL1C_20150912T101403_N0204_R022_T32TPR_20150912T101724'
# sat_folder = '/Users/yw/Downloads'



is_file = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER/test_in_situ.csv'
out_file = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER/test_out.csv'


test = YW_matchups.run(sat_folder, is_file, AC='ACOLITE', out_file=out_file, to_log = True)




'''
# Plot 

from matplotlib import pyplot as plt
plt.imshow(test3, interpolation='nearest')
plt.show()

'''







# Tool package for ACOLITE


import netCDF4 as nc4
import ast
import numpy as np
import datetime as dt



# Get datetime object 
def get_datetime(sat_file):
    with nc4.Dataset(sat_file,"r") as nc:
        
        sensor = nc.getncattr('sensor')

        
        ### full resolution ###
        # datetime_sat_string = nc.getncattr('isodate')
        # format_string = '%Y-%m-%dT%H:%M:%S.%f%z'
        
        # March 20, 2024 update: prioritize inputfile name instead, for S2
        
        try: 
            
            if sensor=='L8_OLI': raise Exception()
            if sensor=='L9_OLI': raise Exception()
            if sensor=='PHR1A': raise Exception()
            if sensor=='PHR1B': raise Exception()
            
            
            inputfile = nc.getncattr('inputfile').split('/')[-1]
            datetime_sat_string = inputfile[11:26]
            format_string = '%Y%m%dT%H%M%S'
            
        except:
            # nearest second 
            datetime_sat_string = nc.getncattr('isodate').split('.', 1)[0]
            format_string = '%Y-%m-%dT%H:%M:%S'
            
        datetime_sat_obj = dt.datetime.strptime(datetime_sat_string, format_string)
    return [datetime_sat_obj, False]
    
    

# Read NC data 

def read_file(sat_file, ACOLITE_L2R, ACOLITE_L1R):
    '''
    sat_file: path to ACOLITE netCDF
    ACOLITE_L2R: Boolean. True: L2R file. False: L2W file. 

    '''
    
    with nc4.Dataset(sat_file,"r") as nc:
        sensor = nc.getncattr('sensor')
        
        # S2 wavelengths 
        if sensor == 'S2A_MSI' or sensor == 'S2B_MSI':
            column_names = ['B1','B2','B3','B4','B5','B6','B7','B8','B8A','B11','B12']
        
            if sensor == 'S2A_MSI':
                bands = [443,492,560,665,704,740,783,833,865,1614,2202] 
            elif sensor == 'S2B_MSI':
                bands = [442,492,559,665,704,739,780,833,864,1610,2186]
        
        # L8 wavelengths
        elif sensor == 'L8_OLI':
            bands = [443, 483, 561, 655, 865, 1609, 2201]
            column_names = ['B1','B2','B3','B4','B5','B6','B7']
        
        # L9
        elif sensor == 'L9_OLI':
            bands = [443, 482, 561, 654, 865, 1608, 2201]
            column_names = ['B1','B2','B3','B4','B5','B6','B7']
        
        
        # Pleiades 
        elif sensor == 'PHR1A' or sensor == 'PHR1B': 
            column_names = ['B1','B2','B3','B4']
            
            if sensor == 'PHR1A':
                bands = [501,561,650,835] 
            elif sensor == 'PHR1B':
                bands = [505,558,663,844]
        
        # band names for extracting values
        if ACOLITE_L2R: 
            band_names = ['rhos_' + str(item) for item in bands]
        elif ACOLITE_L1R:
            band_names = ['rhot_' + str(item) for item in bands]
        else: 
            band_names = ['rhow_' + str(item) for item in bands]

        # dimension
        
        if sensor == 'PHR1A' or sensor == 'PHR1B': 
            global_dims = nc.getncattr('data_dimensions')
            
        else:
            global_dims = nc.getncattr('global_dims')
        
        
        height = global_dims[0] # equivalent to 'lines' before 
        width = global_dims[1]  # equivalent to 'pixels' before 

        # latlon
        lat = nc.variables['lat'][:,:]
        lon = nc.variables['lon'][:,:]

        # to store data
        rhow_data = np.ma.empty([len(bands),int(height),int(width)])

        for k in range(len(bands)):
            varname = band_names[k]
            rhow_data[k] = nc.variables[varname][:,:]
            
        # remove non-water pixels in L2R mode 
        if ACOLITE_L2R or ACOLITE_L1R: 
            # 1600 nm band 
            index_1600 = min(range(len(bands)), key=lambda i: abs(bands[i] - 1600))
            
            # Always use L1R mask 
            # smartHarbour I used L2R
            
            
            # non-water mask 
            if ACOLITE_L1R:
                my_mask = rhow_data[index_1600] > 0.05
                
            # L1R: use mask from L2R
            else:
                rhow_data_L1R = np.ma.empty([len(bands),int(height),int(width)])
                band_names_L1R = ['rhot_' + str(item) for item in bands]

                for k in range(len(bands)):
                    varname = band_names_L1R[k]
                    rhow_data_L1R[k] = nc.variables[varname][:,:]
                    
                my_mask = rhow_data_L1R[index_1600] > 0.05
                
            rhow_data[:, my_mask] = np.nan

    return {'bands': band_names,            # variable names of the bands 
            'sensor': sensor,
            'column_names': column_names,   # column names for final output
            'width': width,                 # width of the data array
            'height': height,               # height of the data array
            'lat': lat,                     # 2d data array: latitude at [y,x]
            'lon': lon,                     # 2d data array: longitude at [y,x]
            'rhow_data': rhow_data}         # 3d data array: reflectance at [band, y, x]
























# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import YW_matchups








# sat_file = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER/test_output_S2_60m.nc'


# sat_folder = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER'

sat_folder = '/Users/yw/Downloads/L8_out'
# sat_folder = '/Users/yw/Downloads'



is_file = '/Users/yw/Local_storage/GLORIA/6_output_L8_noRrs.csv'


out_file = '/Users/yw/Downloads/L8_out/text.csv'


test = YW_matchups.run(sat_folder, is_file, AC='ACOLITE', out_file=out_file, to_log = True)




'''
# Plot 

# Interactive mode
%matplotlib qt


test = sat_data['rhow_data']

test2 = test[0]



from matplotlib import pyplot as plt
plt.imshow(test, interpolation='nearest')
plt.show()

'''




from setuptools import setup, find_packages

with open("readme.md", "r") as fh:
    long_description = fh.read()

setup(
    name='yw_matchups',
    version='2.2.1',
    author='Yulun Wu',
    author_email='yulunwu8@gmail.com',
    description='Find satellite matchups for aquatic remote sensing',
    long_description=long_description,      # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    python_requires='>=3.8',
    install_requires=['pyproj','pandas','numpy','netCDF4']
)
















# Strategy: throw everything here, put them into modules if too long or when necessary 







### Introduction 

# find matchups, mean and median for each of the window sizes 
# sensor in the last column 


### Problem

# POLYMER no sensing time for L8, therefore add a day to the time window 
# should be okay because L8 rarely visits the same place two days in a row





import numpy as np
import sys
import glob, os
import pyproj
import datetime as dt
import pandas as pd
import time
import warnings
from importlib.metadata import version
from scipy.spatial import cKDTree


from . import AC_polymer
from . import AC_acolite
from . import AC_l2gen


# make time column and string an option 



def run(sat_folder=None, is_file=None, AC=None, out_file=None, to_log = True, 
        dt_column='GLORIA_time', dt_string = "%Y-%m-%dT%H:%M", output_sd = False,
        ACOLITE_L2R=False, ACOLITE_L1R=False):
    
    
    ### Settings to be included later 
    time_window_minutes = 720 # 60*24 # 720 
    max_dist = 100
    
    
    # make it adaptive to sensor resolution ??
    # window_size = 3
    window_size = [1, 3, 5, 7]
    
    
    ### Print all metadata/settings and save them in a txt file
    
    if to_log: 
        # Start logging in txt file
        orig_stdout = sys.stdout
        
        log_file = out_file.replace(".csv", ".txt")
        # log_file = 'tmart_log_' + time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime(time.time())) + '.txt'
    
        class Logger:
            def __init__(self, filename):
                self.console = sys.stdout
                self.file = open(filename, 'w')
                self.file.flush()
            def write(self, message):
                self.console.write(message)
                self.file.write(message)
            def flush(self):
                self.console.flush()
                self.file.flush()
    
        sys.stdout = Logger(log_file)
        
    
    # Metadata
    print('YW_matchups version: ' + str(version('YW_matchups')))
    print('System time: ' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time())))
    print('sat_folder: ' + str(sat_folder))
    print('In-situ file: ' + str(is_file))
    print('AC: ' + str(AC))
    print('out_file: ' + str(out_file))
    print('dt_column: ' + str(dt_column))
    print('dt_string: ' + str(dt_string))
    print('output_sd: ' + str(output_sd))
    print('ACOLITE_L2R: ' + str(ACOLITE_L2R))
    print('ACOLITE_L1R: ' + str(ACOLITE_L1R))
    if to_log: print('log_file: ' + str(log_file))
    
    # Settings 
    print('window_size: ' + str(window_size))
    print('time_window_minutes: ' + str(time_window_minutes))
    print('max_dist: ' + str(max_dist))
    
    
    # Read in-situ data 
    df_is = pd.read_csv(is_file)
    dt_is_string = list(df_is.loc[:,dt_column])
    dt_is_obj = [dt.datetime.strptime(dt_is_string[d],dt_string)for d in list(range(len(dt_is_string)))]
    
    
    
    
    ### find all NC files 
    
    # Keywords for .nc files 
    if AC=='POLYMER' or AC=='L2gen': 
        AC_file = '*.nc'
    elif AC=='ACOLITE':
        AC_file = '*L2W.nc'
        if ACOLITE_L2R or ACOLITE_L1R: AC_file = '*L2R.nc'
    else: 
        sys.exit('Unknown AC algorithm')
    
    # Find them in the directory and subdirectory 
    # https://stackoverflow.com/questions/18394147/how-to-do-a-recursive-sub-folder-search-and-return-files-in-a-list
    sat_files = [y for x in os.walk(sat_folder) for y in glob.glob(os.path.join(x[0], AC_file))]
    sat_files.sort()
    
    
    
    # Output 
    df_output = pd.DataFrame()
    
    
    ### Loop through each of the sat_files
    
    # sat_i = 1
    # sat_file = sat_files[sat_i]
    
    for sat_i, sat_file in enumerate(sat_files): 
        
        print('\n{}/{}'.format(sat_i, len(sat_files)))
        print('Processing: {}'.format(sat_file))
        
        tw_minutes = time_window_minutes
        
        
        
        # Read datetime, for different AC
        if AC=='POLYMER': 
            dt_sat_obj, add_a_day = AC_polymer.get_datetime(sat_file)
        elif AC=='ACOLITE':
            dt_sat_obj, add_a_day = AC_acolite.get_datetime(sat_file)
        elif AC=='L2gen':
            dt_sat_obj, add_a_day = AC_l2gen.get_datetime(sat_file)

        # add a day for POLYMER L8 because we don't have the sensing time 
        # this is okay because L8 doesn't visit the same place two days in a row
        if add_a_day: tw_minutes = tw_minutes + 1440


        # Indices of in-situ data that match the sat datetime  
        is_matches = [ind for ind, s in enumerate(dt_is_obj) if abs((dt_sat_obj - s).total_seconds()/60) <= tw_minutes]
        

        # If there are no matching records, skip to next satellite file.
        if len(is_matches)==0:
            print("No matches for this file within the time limit. Next...\n\n")
            continue
        elif len(is_matches) > 1:
            print('Number of matchups: ' + str(len(is_matches)))

        ### Read NC data, for different AC
        if AC=='POLYMER': 
            sat_data = AC_polymer.read_file(sat_file)
        elif AC=='ACOLITE':
            sat_data = AC_acolite.read_file(sat_file, ACOLITE_L2R, ACOLITE_L1R)
        elif AC=='L2gen':
            sat_data = AC_l2gen.read_file(sat_file)
        
        
        sat_lon = sat_data['lon']
        sat_lat = sat_data['lat']
        
        
        sat_lon_max = sat_data['lon'].max()
        sat_lon_min = sat_data['lon'].min()
        sat_lat_max = sat_data['lat'].max()
        sat_lat_min = sat_data['lat'].min() 
     
        
     
        # Calculate the start time
        start = time.time()
        
        print('Building KDT tree...')
        
        
        ny, nx = sat_lon.shape
        # Flatten the longitude and latitude arrays
        sat_lon_flat = sat_lon.ravel()
        sat_lat_flat = sat_lat.ravel()
        # Stack lon and lat to create an array of points
        points = np.column_stack((sat_lon_flat, sat_lat_flat))
        # Build a KDTree
        tree = cKDTree(points)
        
        
        # Calculate the end time and time taken
        end = time.time()
        length = end - start
        
        # Show the results : this can be altered however you like
        print("Took ", length, " seconds\n")
                

        
        ### Loop through indices and row numbers of in-situ matchups
        
        # ind = 11
        # row = is_matches[ind]
        
        for ind, row in enumerate(is_matches):
            print('Processing in-situ match {}, row {} in spreadsheet'.format(ind, row))
            
            ### Identify distances 
            
            in_situ_lat = df_is.loc[row,"lat"]
            in_situ_lon = df_is.loc[row,"lon"]
            
 
            
            
            # Make sure in situ is in sat range 
            if in_situ_lat > sat_lat_max or in_situ_lat < sat_lat_min:
                print("Not in the range. Next...\n")
                continue
            if in_situ_lon > sat_lon_max or in_situ_lon < sat_lon_min:
                print("Not in the range. Next...\n")
                continue
            
            
                    
            # Query the tree for the closest point
            dist, idx = tree.query([in_situ_lon, in_situ_lat])
            # Convert the flat index back to 2D indices
            line, pixel = np.unravel_index(idx, (ny, nx))
            
            
    

            
            '''
            
            # Project X and Y in meters, centred at the in-situ measurement 
            proj = pyproj.Proj(proj="gnom", ellps="WGS84", lat_0=in_situ_lat,
                               lon_0=in_situ_lon, datum="WGS84", units="m")
            x, y = proj(sat_data['lon'], sat_data['lat'])
            in_situ_x, in_situ_y = proj(in_situ_lon, in_situ_lat)
            
            # Calculate distance and indices of the closest 
            dist = np.sqrt((x - in_situ_x)**2 + (y - in_situ_y)**2)
            
            # Pixels within the max distance, True/False
            within_max_dist = dist < max_dist
            
            # Are there any valid pixels left?
            if np.sum(within_max_dist)==0:
                print("No pixels within acceptable distance (",max_dist,"m ). Next...\n")
                continue
            
            ### Find closest and extract values 
            
            indices_closest = np.argwhere(dist == np.min(dist))
            
            line = indices_closest[0][0]
            pixel = indices_closest[0][1]
            
            '''
            
            
            
            # extract the row as a series 
            row_temp = df_is.iloc[row].copy()
            row_temp['image'] = sat_file
            
            # Loop through window size 
            for ws in window_size: 
            
                # Get pixel coordinates of window around the matching point.
                # (v = vertical/line, h = horizontal/pixel)
                v1 = np.int32(line - ((ws-1)/2))
                v2 = np.int32(line + ((ws+1)/2))
                h1 = np.int32(pixel - ((ws-1)/2))
                h2 = np.int32(pixel + ((ws+1)/2))
                
                # Trim edges of window if they're on the edge of the grid.
                line_start,line_end = max(0, v1), min(sat_data['height'], v2)
                # Window pixel start, pixel end.
                pixel_start,pixel_end = max(0, h1), min(sat_data['width'], h2)
                
                # Extract all values in this window
                rhow_data_window = sat_data['rhow_data'][:,line_start:line_end,pixel_start:pixel_end]
                
                # Column names 
                columns_mean    = ['mean_'+item+'_ws'+str(ws) for item in sat_data['column_names']]
                columns_median  = ['median_'+item+'_ws'+str(ws) for item in sat_data['column_names']]
                
                # SD output 
                if output_sd and ws>1: columns_sd  = ['sd_'+item+'_ws'+str(ws) for item in sat_data['column_names']]
                
                
                # Suppress mean/median of empty slice
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                
                    # Mean and median values per band 
                    rhow_data_mean      = np.nanmean(rhow_data_window, axis=(1,2))
                    rhow_data_median    = np.nanmedian(rhow_data_window, axis=(1,2))
                    
                    if output_sd and ws>1: rhow_data_sd = np.nanstd(rhow_data_window, axis=(1,2))
                
                # if rhow_data_mean.mask.all(): 
                #     print('All pixels are masked in the window.')
                #     continue
                
                # Add values to data frame
                row_mean    = pd.Series(data=rhow_data_mean,    index=columns_mean)
                row_median  = pd.Series(data=rhow_data_median,  index=columns_median)
                if output_sd and ws>1: row_sd = pd.Series(data=rhow_data_sd,  index=columns_sd)
                
                
                # if output sd 
                if output_sd and ws>1:
                    row_temp = pd.concat([row_temp, row_mean, row_median, row_sd], axis=0)
                else:
                    row_temp = pd.concat([row_temp, row_mean, row_median], axis=0)

            # Add sensor
            row_temp['sensor'] = sat_data['sensor']

            # Add to dataframe             
            df_output = pd.concat([df_output,row_temp.to_frame().T])
            
    # Stop logging 
    if to_log: sys.stdout = orig_stdout
    
    # Output dataframe 
    df_output.to_csv(out_file, index=False)

    return 0
    
    
    
    
    
    
    
    
    
    
    
    
    













# Go up by 2 directory and import 

import sys
import os.path as path
two_up =  path.abspath(path.join(__file__ ,"../.."))
sys.path.append(two_up)

import YW_matchups








# sat_file = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER/test_output_S2_60m.nc'


# sat_folder = '/Users/yw/Desktop/230613 AEC paper/231130 POLYMER'

# sat_folder = '/Volumes/San2T/GLORIA/ACed/S2_Tmart_ACOLITE/S2A_MSIL1C_20150912T101403_N0204_R022_T32TPR_20150912T101724'
sat_folder = '/Users/yw/Downloads/L8_out/test'



is_file = '/Users/yw/Local_storage/GLORIA/6_output_S2_noRrs.csv'
out_file = '/Users/yw/Desktop/230613 AEC paper/231221 L2gen batch processing/test_out.csv'


test = YW_matchups.run(sat_folder, is_file, AC='L2gen', out_file=out_file, to_log = False)




'''
# Plot 



k=0
varname = band_names[k]
test = nc.groups['geophysical_data'].variables[varname][:]


from matplotlib import pyplot as plt
plt.imshow(test, interpolation='nearest')
plt.show()





'''

